// JavaScript source code
const rateLimit = require("express-rate-limit");
const rateLimitMsg = process.env.RATE_LIMIT;

//Adding rate limit per user of 10 requests within 10 secs
const limiter = rateLimit({
    windowMs: 10 * 1000, // 10 secs
    max: 10, // limit each IP to 10 requests per windowMs
    message: rateLimitMsg,
    headers: true,
});

const dotenv = require('dotenv').config()
const express = require('express');
const bodyParser = require('body-parser');

//All variables declared
var HashMap = require('hashmap');
var app = express();
var lockerTaken = null;
var resJSON;
var fetchLocker = null;
var fetchCustomer = null;
var finalLocker = null;
var finalCustomerId = null;
var mapValue = null;
var key = null;
var arrivedCustomerLocker = null;
var lockerArray = [];

let map = new HashMap();

// get an environment variable
const port = process.env.PORT;
const lockerSpace = process.env.LOCKER_SPACE;
const lockerFull = process.env.LOCKER_FULL;
const lockerEmpty = process.env.LOCKER_EMPTY;
const invalidCustomer = process.env.INVALID_CUSTOMER;

// for parsing application/json
app.use(bodyParser.json()) 

//for rate limiting
app.use(limiter);

// for parsing application/x-www-form-urlencoded
app.use(bodyParser.urlencoded({ extended: true })) 

//Allocate a locker API
app.post('/allocateLocker', (req, res) => {
    console.log("++++++++++++++++++++++++++++Allocating a Locker++++++++++++++++++++++++++++++++");

    //Customer Id is assigned to variable
    fetchCustomer = req.body.customer;
	console.log("========fetchCustomer"+fetchCustomer);

    //creating empty map with null values when first customer arrives
    if (map.size == 0) {
        for (i = 1; i <= lockerSpace; i++) {
            key = "A" + i;
            map.set(key, "");

            //adding empty lockers to array
            lockerArray.push(key);
        }
    }

    if (lockerArray.length != 0) {
        //sorting lockers as per numbers
        lockerArray.sort();

        //reversed array to access the nearest locker to be allocated
        lockerArray.reverse();

        //nearest locker fetched to be allocated to customer
        key = lockerArray[lockerArray.length - 1];

        //check if any customer is allocated with locker
        mapValue = map.get(key);
		
		console.log("========mapValue"+mapValue);

        //checking if there's empty locker space inbetween
        if (mapValue === null || mapValue === "" || mapValue === undefined) {
           
            //search if customer is present at the locker
            arrivedCustomerLocker = map.search(fetchCustomer);

			console.log("========arrivedCustomerLocker"+arrivedCustomerLocker);
			
            //check if the customer is not present then add into nearest locker
            if (arrivedCustomerLocker == null) {
                //if absent add into locker
                map.set(key, fetchCustomer);
				
				console.log("========inside if stmt");

                //assign key to lockerTaken by customer
                lockerTaken = key;

                //remove locker from array as it is filled
                lockerArray.pop();
            }
            else {
                //Alert that customer is already allocated with given locker
                lockerTaken = "The Customer has already been allocated locker " + arrivedCustomerLocker + " !!!"
            }
        }
    }
    else {
        //Alert that lockers are full
        lockerTaken = lockerFull
    }

    //preparing output JSON response
    resJSON = {
        "Locker": lockerTaken
     }

    //sending JSON as API response
    res.json(resJSON);
});

//Deallocate a Customer API
app.post('/deallocateLocker', (req, res) => {
    console.log("++++++++++++++++++++++++++++Deallocating a Customer++++++++++++++++++++++++++++++++");
    
    //Taking locker from input request into variable
    fetchLocker = req.body.locker;

    //check if this locker has been allocated to any customer or not
    mapValue = map.get(fetchLocker);

    //checking if there's empty locker or the invalid locker as input
    if (mapValue != null && mapValue != "" && mapValue != undefined) {

        //Removing customer from that locker
        map.set(fetchLocker, "");

        lockerArray.push(fetchLocker);

        lockerTaken = "The locker " + fetchLocker + " is free for other customer!!!"
    }
    else {
        lockerTaken = "The locker " + fetchLocker + " is either already free or doesnt exists!!!"
    }

    //Preparing output JSON response
    resJSON = {
        "Locker": lockerTaken
    }

    //sending JSON as API response
    res.json(resJSON);
});

//Get Locker or Customer Information API
app.get('/getlockerandcustomerinfo', (req, res) => {
    console.log("++++++++++++++++++++++++++++Get locker or Customer Info++++++++++++++++++++++++++++++++");

    //Taking locker or Customer Id from input request into variables
    fetchLocker = req.body.locker;
    fetchCustomer = req.body.customer;

    //If input locker field is not empty take it as input to fetch related Customer info
    if (fetchLocker != null && fetchLocker != "") {
        finalLocker = fetchLocker;

        //Check if locker system contains locker as key
        if (map.has(fetchLocker)) {

            //Fetch related customer who is allocated this locker
            finalCustomerId = map.get(fetchLocker);
        }
        else {
            finalCustomerId = lockerEmpty
        }
    }

    //If input Customer Id field is not empty take it as input to fetch related locker number
    else if (fetchCustomer != null && fetchCustomer != "") {
        finalCustomerId = fetchCustomer;

        //search locker system for customer Id and return the locker
        finalLocker = map.search(fetchCustomer);
        if (finalLocker == null || finalLocker == ""){
            finalLocker = ""
            finalCustomerId = invalidCustomer
        }
    }
    else {
        finalCustomerId = invalidCustomer
    }
    
    resJSON = {
        "Locker": finalLocker,
        "CustomerId": finalCustomerId
    }

    //sending JSON as API response
    res.json(resJSON);
});

app.listen(port, () => console.log(`Listening on port ${port}...........`));

console.log(port);